//licensed under a Creative Commons Attribution Non-Commercial Share Alike license
//code by Nervous System http://n-e-r-v-o-u-s.com

import controlP5.*;

import peasy.org.apache.commons.math.*;
import peasy.*;
import peasy.org.apache.commons.math.geometry.*;

import processing.opengl.*;
import javax.media.opengl.*  ;

import processing.video.*;

import toxi.geom.*;
import toxi.geom.mesh.*;
import toxi.volume.*;

TriangleMesh mesh;
String meshFile = "sphereSmall3.stl";

float u[]; //same size as mesh vertices
float v[];
float uNext[];

float reactionRate = 1.0;
float diffusionRateU = .1;
float diffusionRateV = .025;
float deltaT = 1.0;

float F = .06;
float k = .06;

float minU, maxU;

PeasyCam cam;
void setup() {
  size(500,500,OPENGL);
  cam = new PeasyCam(this, 0,0,0,100);
  loadMesh();
  noStroke();
  
  maxU = -999;
  minU = 999;
  for(int i=0;i<u.length;++i) {
    u[i] = .5 +random(-.01,.01);
    v[i] = .25 + random(-.01,.01);
    if(u[i] > maxU) maxU = u[i];
    if(u[i] < minU) minU = u[i];
  }
}

//open stl file 'meshFile'
void loadMesh() {
  STLReader reader = new STLReader();
  mesh = (TriangleMesh) reader.loadBinary(dataPath(meshFile), STLReader.TRIANGLEMESH);
  mesh.scale(10);
  u = new float[mesh.getNumVertices()];
  v = new float[mesh.getNumVertices()];
  uNext = new float[mesh.getNumVertices()];
}

void draw() {
  background(255);
  diffusionU();
  diffusionV();
  reaction();
  
  drawMesh();
}

void drawMesh() {
  beginShape(TRIANGLES);
  int numFaces = mesh.getNumFaces();
  for(int i=0;i<numFaces;++i) {
    Face f = (Face) mesh.faces.get(i);
    fill((u[f.a.id]-minU)/(maxU-minU)*255);
    vertex(f.a.x, f.a.y, f.a.z);

    fill((u[f.b.id]-minU)/(maxU-minU)*255);
    vertex(f.b.x, f.b.y, f.b.z);

    fill((u[f.c.id]-minU)/(maxU-minU)*255);
    vertex(f.c.x, f.c.y, f.c.z);
  }
  endShape();
}


/*
void diffusionU() {
  Vertex v1,v2,v3;
  int i1,i2,i3;
  float area;
  Vec3D cross,d1,d2,d3;
  float sinMag, cos1,cos2,cos3, sin1,sin2,sin3;
  float dist12;
  float dist13;
  float dist23;
  for(Face face : mesh.faces) {
    v1 = face.a;
    v2 = face.b;
    v3 = face.c;
    i1 = v1.id;
    i2 = v2.id;
    i3 = v3.id;
    d1 = v2.sub(v1);
    d2 = v3.sub(v1);
    d3 = v3.sub(v2);
    dist12 = d1.magnitude();
    dist13 = d2.magnitude();
    dist23 = d3.magnitude();
    
    
    cross = d1.cross(d2);
    sinMag = cross.magnitude();

    sin1 = sinMag/(dist12*dist13);
    sin2 = sinMag/(dist12*dist23);
    sin3 = sinMag/(dist13*dist23);
    
    cos1 = d1.x*d2.x+d1.y*d2.y+d1.z*d2.z;
    cos2 = -d1.x*d3.x-d1.y*d3.y-d1.z*d3.z;
    cos3 = d3.x*d2.x+d3.y*d2.y+d3.z*d2.z;
    cos1 /= dist12*dist13;
    cos2 /= dist12*dist23;
    cos3 /= dist13*dist23;
    
    uNext[i1] += deltaT*diffusionRateU*u[i2]*sin1/((cos1+1)*dist12);
    uNext[i1] += deltaT*diffusionRateU*u[i3]*sin1/((cos1+1)*dist13);
    uNext[i1] += deltaT*diffusionRateU*u[i1]*-sin1/((cos1+1)*dist12);
    uNext[i1] += deltaT*diffusionRateU*u[i1]*-sin1/((cos1+1)*dist13);

    uNext[i2] += deltaT*diffusionRateU*u[i1]*sin2/((cos2+1)*dist12);
    uNext[i2] += deltaT*diffusionRateU*u[i3]*sin2/((cos2+1)*dist13);
    uNext[i2] += deltaT*diffusionRateU*u[i2]*-sin2/((cos2+1)*dist12);
    uNext[i2] += deltaT*diffusionRateU*u[i2]*-sin2/((cos2+1)*dist23);

    uNext[i3] += deltaT*diffusionRateU*u[i1]*sin3/((cos3+1)*dist13);
    uNext[i3] += deltaT*diffusionRateU*u[i2]*sin3/((cos3+1)*dist23);
    uNext[i3] += deltaT*diffusionRateU*u[i3]*-sin3/((cos3+1)*dist13);
    uNext[i3] += deltaT*diffusionRateU*u[i3]*-sin3/((cos3+1)*dist23);
  }

  float[] temp = u;
  u = uNext;
  uNext = temp;

}
*/


void diffusionU() {
  int numFaces = mesh.getNumFaces();
  for(int i=0;i<u.length;++i) {
    uNext[i] = u[i];
  }
  for(int i = 0;i<numFaces;++i) {
    Face f = (Face) mesh.faces.get(i);
    //between vertex a + b
    //access chemicals index through vertex id
    float diffusion = deltaT*diffusionRateU*(u[f.a.id]-u[f.b.id]);
    uNext[f.b.id] += diffusion;
    uNext[f.a.id] += -diffusion;
    //b to c
    diffusion = deltaT*diffusionRateU*(u[f.b.id]-u[f.c.id]);
    uNext[f.c.id] += diffusion;
    uNext[f.b.id] += -diffusion;
    //c to a
    diffusion = deltaT*diffusionRateU*(u[f.c.id]-u[f.a.id]);
    uNext[f.a.id] += diffusion;
    uNext[f.c.id] += -diffusion;
  }
  float[] temp = u;
  u = uNext;
  uNext = temp;
}

void diffusionV() {
  int numFaces = mesh.getNumFaces();
  for(int i=0;i<u.length;++i) {
    uNext[i] = v[i];
  }

  for(int i = 0;i<numFaces;++i) {
    Face f = (Face) mesh.faces.get(i);
    //between vertex a + b
    //access chemicals index through vertex id
    float diffusion = deltaT*diffusionRateV*(v[f.a.id]-v[f.b.id]);
    uNext[f.b.id] +=diffusion;
    uNext[f.a.id] +=-diffusion;
    //b to c
    diffusion = deltaT*diffusionRateV*(v[f.b.id]-v[f.c.id]);
    uNext[f.c.id] +=diffusion;
    uNext[f.b.id] +=-diffusion;
    //c to a
    diffusion = deltaT*diffusionRateV*(v[f.c.id]-v[f.a.id]);
    uNext[f.a.id] += diffusion;
    uNext[f.c.id] += -diffusion;
  }
  float[] temp = v;
  v = uNext;
  uNext = temp;
}

void reaction() {
  minU = 999;
  maxU = -999;
  for(int i=0;i<u.length;++i) {
    float uVal = u[i];
    u[i] = u[i]+deltaT*reactionU(uVal, v[i], F, k);
    v[i] = v[i]+deltaT*reactionV(uVal, v[i], F, k);
    if(u[i] > maxU) maxU = u[i];
    if(u[i] < minU) minU = u[i];
  }
}



//Grey-Scott
float reactionU(float aVal, float bVal, float FVal, float kVal) {
  return reactionRate*(-aVal*bVal*bVal+F*(1-aVal));
}

float reactionV(float aVal, float bVal, float FVal, float kVal) {
  return reactionRate*(aVal*bVal*bVal-(F+k)*bVal);
}



